/*
 *   Hp-9121.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   HP9121 floppy disk as ADDRESS 0 simulation based on behaviors from
//
//   "HP flexible disc drive command set Appendix A" from an HP9121 manual, amigo protocol
//   hpfddcs.pdf found on the WWW
//   "Original 320 rom listing from HP"
//   ROM320_P1.pdf from Andr Koppel
//

#include "StdAfx.h"
#include "HP8x.h"
#include "kml.h"
#include "io.h"								// I/O definitions

#define HP9121ADR 0x0						// HP-IB address

//#define DEBUG_HP9121						// debug flag

#if defined DEBUG_HP9121
static WORD HP9121_ADR[] = {				// states of automata
	0, 1000, 1100, 1400, 
	2000,
	10000, 10500, 11000, 11500,
	12000, 12500, 13000, 13500,
	14000, 14500, 15000, 15500,
	16000, 16500, 17000, 17500,
	18000, 18500, 19000, 
	20000, 20500, 21000, 21500,
	22000, 22500
};

static TCHAR *HP9121_TXT[] ={				// labels of states of automata 
	_T("Idle"), _T("MSA received"), _T("MLA received"), _T("MTA received"),
	_T("Identify"),
	_T("Receive data"), _T("Cold load read"), _T("Seek"), _T("Request status"),
	_T("Verify"), _T("Initialize"), _T("Request logical address"), _T("End"),
	_T("buffered write ?"), _T("Buffered read ?"), _T("unbuffered read verify"), _T("request physical address"),
	_T("Send wear"), _T("Format"), _T("Download"), _T("HP-300 clear"),
	_T("HP-IB-CRC L"), _T("Write loopback record"), _T("Initiate selftest"), 
	_T("Send data"), _T("Send status or address"), _T("DSJ"), _T("HP-IB-CRC T"),
	_T("Read loopback record"), _T("Read selftest")
};

#define NLAB (sizeof(HP9121_ADR)/sizeof(HP9121_ADR[0]))

static TCHAR *HP9121_LAB(WORD d)			// find a label for the state d
{
	WORD i;

	for(i = 0; i < NLAB; i++)
		if (HP9121_ADR[i] == d)
			return HP9121_TXT[i];
	return NULL;
}
#endif

#define MIN(a, b) ((a<b)?a:b)

// hp9121 variables

#define DISK_BYTES ((hp9121_config.heads * hp9121_config. cylinders * hp9121_config.sectors) << 8)
													// 0x01,0x04 ->    2,       33,     32
													// 0x00,0x81 ->    2,       75,     60
													// 0x01,0x06 ->    4,      153,    124
													//              head, cylinder, sect/cylinder (include head)

typedef struct 
{
	BYTE heads;
	BYTE cylinders;
	BYTE sectors;
} HP9121_CONFIG;

static LPBYTE Disk[10] = {NULL, NULL, NULL, NULL,
						  NULL, NULL, NULL, NULL,
					 	  NULL, NULL};			// disks for units data
static WORD Status[10] = {0,0,0,0,0,0,0,0,0,0};		// status of unit  
													// x x x x x x x x x x x x x x x x
													//                             | +- disk present
													//                             +--- read-only (not used)
static WORD st9121 = 0;								// state of hp9121 controller
static BOOL hp9121_talk = FALSE;					// MTA received ?
static BOOL hp9121_listen = FALSE;					// MLA received ?
static BOOL hp9121_untalk = FALSE;					// previous command was UNTALK ?
static BYTE hp9121_s1[2] = {0, 0};					// status1
static BYTE hp9121_s2[2] = {0, 0};					// status2
static BYTE hp9121_dsj = 0x00;						// DSJ of controler
static BYTE hp9121_head[10] = {0,0,0,0};			// current head
static WORD hp9121_cylinder[10] = {0,0,0,0};		// current cylinder
static BYTE hp9121_sector[10] = {0,0,0,0};			// current track
static BYTE hp9121_unit = 0;						// last addressed unit
static BYTE hp9121_type[2] = {1,4};					// kind of unit
static HP9121_CONFIG hp9121_config = {2, 33, 16};	// config of unit (head, cylinder, sect/cylinder)
static DWORD hp9121_adr[10] = {0,0,0,0};			// adresse of byte in disk
static BYTE hc_9121[0x200];							// stack of commands
static WORD hc_9121_hi = 0;							// hi mark
static WORD hc_9121_lo = 0;							// lo mark
static BYTE	hd_9121[0x200];							// stack of data in
static WORD hd_9121_hi = 0;							// hi mark
static WORD hd_9121_lo = 0;							// lo mark
static BYTE message[4];								// message status or address
static BOOL drawoff3 = TRUE;
static BOOL drawon3 = FALSE;

BOOL hp9121_ppol_e = TRUE;							// parallel poll enabled

//################
//#
//#    Low level subroutines
//#
//################

static VOID hp9121_raj_adr(BYTE u)
{
	while (hp9121_sector[u] >= hp9121_config.sectors)
		hp9121_sector[u] -= hp9121_config.sectors;
	while (hp9121_head[u] >= hp9121_config.heads)
		hp9121_head[u] -= hp9121_config.heads;
	while (hp9121_cylinder[u] >= hp9121_config.cylinders)
		hp9121_cylinder[u] -= hp9121_config.cylinders;
	hp9121_adr[u] = (hp9121_sector[u] + 
					 hp9121_head[u] * hp9121_config.sectors + 
					 hp9121_cylinder[u] * hp9121_config.heads * hp9121_config.sectors
					) * 256;
}

static VOID hp9121_inc_adr(BYTE u)
{
	hp9121_sector[u]++;
	if (hp9121_sector[u] >= hp9121_config.sectors)
	{
		hp9121_sector[u] = 0;
		hp9121_head[u]++;
		if (hp9121_head[u] >= hp9121_config.heads)
		{
			hp9121_head[u] = 0;
			hp9121_cylinder[u]++;
			if (hp9121_cylinder[u] >= hp9121_config.cylinders)
				hp9121_cylinder[u] = 0;	// wrap
		}
	}
	hp9121_raj_adr(u);
}

static BOOL hp9121_pop_c(BYTE *c)
{
	if (hc_9121_hi == hc_9121_lo)
		return FALSE;
	*c = hc_9121[hc_9121_lo++];
	hc_9121_lo &= 0x1FF;
	return TRUE;
}

static BOOL hp9121_pop_d(BYTE *c)
{
	if (hd_9121_hi == hd_9121_lo)
		return FALSE;
	*c = hd_9121[hd_9121_lo++];
	hd_9121_lo &= 0x1FF;
	return TRUE;
}

BOOL hp9121_push_c(BYTE addr, BYTE c)	// push on stack hc_9121
{
	if (addr != HP9121ADR)
		return FALSE;
	hc_9121[hc_9121_hi++] = c;
	hc_9121_hi &= 0x1FF;
	_ASSERT(hc_9121_hi != hc_9121_lo);
	return TRUE;
}

BOOL hp9121_push_d(BYTE addr, BYTE d)	// push on stack hd_9121
{
	
	if (addr != HP9121ADR || !hp9121_listen)
		return FALSE;
	hd_9121[hd_9121_hi++] = d;
	hd_9121_hi &= 0x1FF;
	_ASSERT(hd_9121_hi != hd_9121_lo);
	return TRUE;
}

//################
//#
//#    Public functions
//#
//################

//
//   FSA to simulate the disc controller (Amigo protocol)
//

VOID DoHP9121(VOID)
{
#if defined DEBUG_HP9121
WORD hp9121st;
#endif

BYTE c;
WORD d;

	#if defined DEBUG_HP9121
	hp9121st = st9121;
	#endif

	switch (st9121)
	{
	// IDLE STATE
		case 0:
			Chipset.annun &= ~0x00000002;
			DrawAnnunciator(2, FALSE);
			st9121++;
			break;
		case 1:
			if (!drawoff3)
			{
				if ((h_out_hi == h_out_lo) && (hd_9121_hi == hd_9121_lo)) 
				{
					Chipset.annun &= ~(1 << (2 + hp9121_unit*2));
					DrawAnnunciator(3 + hp9121_unit*2, FALSE);
					drawoff3 = TRUE;
					drawon3 = FALSE;
				}
			}
			if (!drawon3)
			{
				if ((h_out_hi != h_out_lo) || (hd_9121_hi != hd_9121_lo)) 
				{
					Chipset.annun |= (1 << (2 + hp9121_unit*2));
					DrawAnnunciator(3 + hp9121_unit*2, TRUE);
					drawoff3 = FALSE;
					drawon3 = TRUE;
				}
			}
			if (hc_9121_hi != hc_9121_lo)
			{
				hp9121_pop_c(&c);
				c &= 0x7F;						// remove parity
				switch(c)
				{
					case (0x60 + HP9121ADR):	// my secondary address
						if (hp9121_untalk)
						{
							Chipset.annun |= 0x00000002;
							DrawAnnunciator(2, TRUE);
							st9121 = 1000;
						}
						break;
					case (0x20 + HP9121ADR):	// my listen address
						hp9121_listen = TRUE;
						Chipset.annun |= 0x00000002;
						DrawAnnunciator(2, TRUE);
						hp9121_untalk = FALSE;
						st9121 = 1100;
						break;
					case (0x40 + HP9121ADR):	// my talk address
						hp9121_talk = TRUE;
						Chipset.annun |= 0x00000002;
						DrawAnnunciator(2, TRUE);
						hp9121_untalk = FALSE;
						st9121 = 1400;
						break;
					case 0x3F:					// unlisten
						hp9121_listen = FALSE;
						hp9121_untalk = FALSE;
						break;
					case 0x5F:					// untalk
						hp9121_talk = FALSE;
						hp9121_untalk = TRUE;
						break;
					default:					// unknown
						hp9121_untalk = FALSE;
						break;
				}
			}
			break;
	// receive my secondary address
		case 1000:
			if (hp9121_untalk)
				st9121 = 2000;
			else
				st9121 = 0;
			hp9121_untalk = FALSE;
			break;
	// receive my listen address
		case 1100:
			if (hp9121_pop_c(&c))
			{
				c &= 0x7F;						// remove parity
				switch(c)
				{
					case 0x60:					// secondary 0x00	: Receive data
						hp9121_ppol_e = FALSE;
						st9121 = 10000;
						break;
					case 0x68:					// secondary 0x08	: misc
						st9121 = 1200;
						break;
					case 0x69:					// secondary 0x09   : Buffered write ?
						hp9121_ppol_e = FALSE;
						st9121 = 14000;
						break;
					case 0x6A:					// secondary 0x0A   : Buffered read ?
						hp9121_ppol_e = FALSE;
						st9121 = 14500;
						break;
					case 0x6C:					// secondary 0x0C   : misc
						st9121 = 1300;
						break;
					case 0x6F:					// secondary 0x0F   : Download
						hp9121_ppol_e = FALSE;
						st9121 = 17000;
						break;
					case 0x70:					// secondary 0x10   : HP-300 clear
						hp9121_ppol_e = FALSE;
						st9121 = 17500;
						break;
					case 0x71:					// secondary 0x11   : HP-IB CRC
						hp9121_ppol_e = FALSE;
						st9121 = 18000;
						break;
					case 0x7E:					// secondary 0x1E   : Write loopback record
						hp9121_ppol_e = FALSE;
						st9121 = 18500;
						break;
					case 0x7F:					// secondary 0x1F   : Initiate selftest
						hp9121_ppol_e = FALSE;
						st9121 = 19000;
						break;
					default:
						_ASSERT(0);
						st9121 = 0;				// restart
						break;
				}
			}
			break;
	// MLA & secondary 08 
		case 1200:
			hp9121_ppol_e = FALSE;
			st9121++;
			break;
		case 1201:
			if (hp9121_pop_d(&c))
			{
				switch (c)
				{
					case 0x00:					// op 0x00 : Cold load read
						st9121 = 10500;
						break;
					case 0x02:					// op 0x02 : Seek
						st9121 = 11000;
						break;
					case 0x03:					// op 0x03 : Request status
						st9121 = 11500;
						break;
					case 0x07:					// op 0x07 : Verify
						st9121 = 12000;
						break;
					case 0x0B:					// op 0x0B : Initialize
						st9121 = 12500;
						break;
					case 0x14:					// op 0x14 : Request logical address
						st9121 = 13000;
						break;
					case 0x15:					// op 0x15 : End
						st9121 = 13500;
						break;
					default:
						_ASSERT(0);
						hp9121_ppol_e = FALSE;
						st9121 = 0;				// restart
						break;
				}
			}
			break;
	// MLA & secondary 0C 
		case 1300:
			hp9121_ppol_e = FALSE;
			st9121++;
			break;
		case 1301:
			if (hp9121_pop_d(&c))
			{
				switch (c)
				{
					case 0x05:					// op 0x05 : Unbuffered read verify
						st9121 = 15000;
						break;
					case 0x14:					// op 0x14 : Request physical address
						st9121 = 15500;
						break;
					case 0x16:					// op 0x16 : Send wear
						st9121 = 16000;
						break;
					case 0x18:					// op 0x18 : Format
						st9121 = 16500;
						break;
					default:
						_ASSERT(0);
						st9121 = 0;				// restart
						break;
				}
			}
			break;
	// receive my talk address
		case 1400:
			if (hp9121_pop_c(&c))
			{
				c &= 0x7F;						// remove parity
				switch(c)
				{
					case 0x60:					// secondary 0x00	: Send data
						hp9121_ppol_e = FALSE;
						st9121 = 20000;
						break;
					case 0x68:					// secondary 0x08	: Send status or address
						hp9121_ppol_e = FALSE;
						st9121 = 20500;
						break;
					case 0x70:					// secondary 0x10   : DSJ
						hp9121_ppol_e = FALSE;
						st9121 = 21000;
						break;
					case 0x71:					// secondary 0x11   : HP-IB CRC
						hp9121_ppol_e = FALSE;
						st9121 = 21500;
						break;
					case 0x7E:					// secondary 0x1E   : Read loopback record
						hp9121_ppol_e = FALSE;
						st9121 = 22000;
						break;
					case 0x7F:					// secondary 0x1F   : Read selftest
						hp9121_ppol_e = FALSE;
						st9121 = 22500;
						break;
					default:
						_ASSERT(0);
						st9121 = 0;				// restart
						break;
				}
			}
			break;
	// Identify command
		case 2000:
			h_push(hp9121_type[0], 0x00);		// even if talk not enable ....
			h_push(hp9121_type[1], 0x80);
			st9121 = 0;
			break;
	// MLA & secondary 00 : receive data
		case 10000:
			if (hp9121_listen)
			{
				if (hp9121_pop_d(&c))
				{
					Disk[hp9121_unit][hp9121_adr[hp9121_unit]++] = c;
					if ((hp9121_adr[hp9121_unit] & 0xFF) == 0x00)
					{
						hp9121_inc_adr(hp9121_unit);
						hp9121_ppol_e = TRUE;
						st9121 = 0;
					}
				}
			}
			break;
	// MLA & secondary 08 & 0x00 : cold load read
		case 10500:
			_ASSERT(0);
			break;
	// MLA & secondary 08 & 0x02 : seek
		case 11000:
			if (hp9121_pop_d(&c))
			{
				hp9121_unit = MIN(9, c);
				st9121++;
			}
		case 11001:
			if (hp9121_pop_d(&c))
			{
				hp9121_cylinder[hp9121_unit] = c << 8;
				st9121++;
			}
			break;
		case 11002:
			if (hp9121_pop_d(&c))
			{
				hp9121_cylinder[hp9121_unit] |= c;
				st9121++;
			}
			break;
		case 11003:
			if (hp9121_pop_d(&c))
			{
				hp9121_head[hp9121_unit] = c;
				st9121++;
			}
			break;
		case 11004:
			if (hp9121_pop_d(&c))
			{
				hp9121_sector[hp9121_unit] = c;
				hp9121_raj_adr(hp9121_unit);
				if ((Status[hp9121_unit] & 0x0001) == 0x0001)	//  disk
				{
					hp9121_s1[0] = 0;
					hp9121_s1[1] = 0;
					hp9121_dsj = 0;
				}
				else
				{
					hp9121_dsj = 1;
					hp9121_s1[0] = 0x1F;					// 00dSSSSS : 00011111
					hp9121_s1[1] = hp9121_unit;				// unit
					hp9121_s2[0] = 0x2D;					// *TTTTR     00101101
					hp9121_s2[1] = 0x87;					// AW/EFCSS   10000111
				}
				hp9121_ppol_e = TRUE;
				st9121 = 0;
			}
			break;
	// MLA & secondary 08 & 0x03 : request status
		case 11500:
			if (hp9121_pop_d(&c))
			{
				c = MIN(9, c);
				message[0] = hp9121_s1[0];
				message[1] = hp9121_s1[1];
//				_ASSERT(c == hp9121_s1[1]);
				message[2] = hp9121_s2[0];
				message[3] = hp9121_s2[1];
				hp9121_s2[1] &= 0x63;				// clear bits
				hp9121_dsj = 0;						// DSJ = 0
				hp9121_ppol_e = TRUE;
				st9121 = 0;
			}
			break;
	// MLA & secondary 08 & 0x07 : verify (stop on last good cylinder)
		case 12000:							// fake it
			if (hp9121_pop_d(&c))				// unit
			{
				hp9121_unit = MIN(9, c);
				st9121++;
			}
			break;
		case 12001:					
			if (hp9121_pop_d(&c))				// sector high
			{
				st9121++;
			}
			break;
		case 12002:					
			if (hp9121_pop_d(&c))				// sector low
			{
				hp9121_sector[hp9121_unit] = 0;
				hp9121_head[hp9121_unit] = 0;
				hp9121_cylinder[hp9121_unit] = hp9121_config.cylinders;		// last cylinder
//				hp9121_raj_adr(hp9121_unit);
				if ((Status[hp9121_unit] & 0x0001) == 0x0001)		// disk
				{
					hp9121_s1[0] = 0;
					hp9121_s1[1] = 0;
					hp9121_dsj = 0;
				}
				else												// no disk
				{
					hp9121_dsj = 1;
					hp9121_s1[0] = 0x13;					// 00dSSSSS : 00010011
					hp9121_s1[1] = hp9121_unit;				// unit
					hp9121_s2[0] = 0x0D;					// *TTTTR     00001101
					hp9121_s2[1] = 0x03;					// AW/EFCSS   00000011
				}
				hp9121_ppol_e = TRUE;
				st9121=0;
			}
			break;
	// MLA & secondary 08 & 0x0B : initialize
		case 12500:
			_ASSERT(0);
			break;
	// MLA & secondary 08 & 0x14 : request logical address
		case 13000:							// dummy byte
			if (hp9121_pop_d(&c))
			{
				st9121++;
			}
			break;
		case 13001:
			message[0] = hp9121_cylinder[hp9121_unit] >> 8;
			message[1] = hp9121_cylinder[hp9121_unit] & 0xFF;
			message[2] = hp9121_head[hp9121_unit];
			message[3] = hp9121_sector[hp9121_unit];
			hp9121_ppol_e = TRUE;
			st9121 = 0;
			break;
	// MLA & secondary 08 & 0x15 : end
		case 13500:
			_ASSERT(0);
			break;
	// MLA & secondary 09 : buffered write ?
		case 14000:
			if (hp9121_pop_d(&c))
			{
				if (c == 0x08)
					st9121++;
				else
				{
					hp9121_ppol_e = TRUE;
					st9121 = 0;
				}
			}
			break;
		case 14001:
			if (hp9121_pop_d(&c))
			{
				hp9121_unit = MIN(9, c);
				if ((Status[hp9121_unit] & 0x0001) == 0x0001)		// disk
				{
					hp9121_s1[0] = 0;
					hp9121_s1[1] = 0;
					hp9121_dsj = 0;
				}
				else												// no disk
				{
					hp9121_dsj = 1;
					hp9121_s1[0] = 0x13;					// 00dSSSSS : 00010011
					hp9121_s1[1] = hp9121_unit;				// unit
					hp9121_s2[0] = 0x0D;					// *TTTTR     00001101
					hp9121_s2[1] = 0x03;					// AW/EFCSS   00000011
				}
				hp9121_ppol_e = TRUE;
				st9121 = 0;
			}
			break;
	// MLA & secondary 0A : buffered read ?
		case 14500:
			if (hp9121_pop_d(&c))
			{
				if (c == 0x05)
					st9121++;
				else
				{
					hp9121_ppol_e = TRUE;
					st9121 = 0;
				}
			}
			break;
		case 14501:
			if (hp9121_pop_d(&c))
			{
				hp9121_unit = MIN(9, c);
				if ((Status[hp9121_unit] & 0x0001) == 0x0001)		// disk
				{
					hp9121_s1[0] = 0;
					hp9121_s1[1] = 0;
					hp9121_dsj = 0;
				}
				else												// no disk
				{
					hp9121_dsj = 1;
					hp9121_s1[0] = 0x13;					// 00dSSSSS : 00010011
					hp9121_s1[1] = hp9121_unit;				// unit
					hp9121_s2[0] = 0x0D;					// *TTTTR     00001101
					hp9121_s2[1] = 0x03;					// AW/EFCSS   00000011
				}
				hp9121_ppol_e = TRUE;
				st9121 = 0;
			}
			break;
	// MLA & secondary 0C & 0x05 : unbuffered read verify
		case 15000:
			_ASSERT(0);
			break;
	// MLA & secondary 0C & 0x14 : request physical address
		case 15500:
			_ASSERT(0);
			message[0] = hp9121_cylinder[hp9121_unit] >> 8;
			message[1] = hp9121_cylinder[hp9121_unit] & 0xFF;
			message[2] = hp9121_head[hp9121_unit];
			message[3] = hp9121_sector[hp9121_unit];
			hp9121_ppol_e = TRUE;
			st9121 = 0;
			break;
	// MLA & secondary 0C & 0x16 : send wear
		case 16000:
			_ASSERT(0);
			break;
	// MLA & secondary 0C & 0x18 : format
		case 16500:							// unit
			if (hp9121_pop_d(&c))
			{
				hp9121_unit = MIN(9, c);
				st9121++;
			}
			break;
		case 16501:							// type
			if (hp9121_pop_d(&c))
			{
				st9121++;
			}
			break;
		case 16502:							// interleave
			if (hp9121_pop_d(&c))
			{
				st9121++;
			}
			break;
		case 16503:							// data
			if (hp9121_pop_d(&c))
			{
				if ((Status[hp9121_unit] & 0x0001) == 0x0001)		// disk
				{
					FillMemory(Disk[hp9121_unit], DISK_BYTES, c);
					hp9121_s1[0] = 0;
					hp9121_s1[1] = 0;
					hp9121_dsj = 0;
				}
				else												// no disk
				{
					hp9121_dsj = 1;
					hp9121_s1[0] = 0x13;					// 00dSSSSS : 00010011
					hp9121_s1[1] = hp9121_unit;				// unit
					hp9121_s2[0] = 0x0D;					// *TTTTR     00001101
					hp9121_s2[1] = 0x03;					// AW/EFCSS   00000011
				}
				hp9121_ppol_e = TRUE;
				st9121 = 0;
			}
			break;
	// MLA & secondary 0F : download
		case 17000:
			_ASSERT(0);
			break;
	// MLA & secondary 10 : HP-300 clear
		case 17500:
			_ASSERT(0);
			break;
	// MLA & secondary 11 : HP-IB CRC
		case 18000:
			_ASSERT(0);
			break;
	// MLA & secondary 1E : write loopback record
		case 18500:
			_ASSERT(0);
			break;
	// MLA & secondary 1F : initiate selftest
		case 19000:
			_ASSERT(0);
			break;
	// MTA & secondary 00 : send data
		case 20000:
			if (hp9121_talk)
			{
				d = 256;
				while (d > 0)
				{
					h_push(Disk[hp9121_unit][hp9121_adr[hp9121_unit]++], 0x00);
					d--;
				}
			}
			hp9121_inc_adr(hp9121_unit);
			hp9121_ppol_e = TRUE;
			st9121 = 0;
			break;
	// MTA & secondary 08 : send status or address
		case 20500:
			if (hp9121_talk)
			{
				h_push(message[0], 0x00);
				h_push(message[1], 0x00);
				h_push(message[2], 0x00);
				h_push(message[3], 0x80);
			}
			hp9121_ppol_e = TRUE;
			st9121 = 0;
			break;
	// MTA & secondary 10 : DSJ
		case 21000:
			if (hp9121_talk)
				h_push(hp9121_dsj, 0x80);		// send DSJ = 0 and EOI
			if (hp9121_dsj == 2)
				hp9121_dsj = 0;
			st9121 = 0;
			break;
	// MTA & secondary 11 : HP-IB CRC
		case 21500:
			_ASSERT(0);
			break;
	// MTA & secondary 1E : read loopback record
		case 22000:
			_ASSERT(0);
			break;
	// MTA & secondary 1E : read selftest
		case 22500:
			_ASSERT(0);
			break;
		default:
			_ASSERT(0);
			break;
	}
	#if defined DEBUG_HP9121
	{
		TCHAR buffer[256];
		TCHAR *TXT = HP9121_LAB((WORD)(st9121));
		int k;
		unsigned long l;

		if (hp9121st != st9121)
		{
			if (TXT == NULL)
				k = wsprintf(buffer,_T("%04X: HP9121 : %u->%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), hp9121st, st9121);
			else
				k = wsprintf(buffer,_T("%04X: HP9121 : %u->%u ( %s )\n"), Chipset.R[4]|(Chipset.R[5]<<8), hp9121st, st9121, TXT);
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
		}
	}
	#endif
}

BOOL hp9121_load(BYTE unit, LPCTSTR szFilename)
{
	HANDLE  hDiskFile = NULL;
	DWORD dwFileSize, dwRead;
	LPBYTE pDisk = NULL;

	hDiskFile = CreateFile(szFilename,
						   GENERIC_READ,
						   FILE_SHARE_READ,
						   NULL,
						   OPEN_EXISTING,
						   FILE_FLAG_SEQUENTIAL_SCAN,
						   NULL);
	if (hDiskFile == INVALID_HANDLE_VALUE)
	{
		hDiskFile = NULL;
		return FALSE;
	}
	dwFileSize = GetFileSize(hDiskFile, NULL);
	if (dwFileSize > (DWORD)(DISK_BYTES))
	{ // file is too big.
		CloseHandle(hDiskFile);
		return FALSE;
	}

	pDisk = HeapAlloc(hHeap, HEAP_ZERO_MEMORY, DISK_BYTES);
	if (pDisk == NULL)
	{
		CloseHandle(hDiskFile);
		return FALSE;
	}

	// load file content)
	ReadFile(hDiskFile, pDisk, dwFileSize, &dwRead,NULL);
	CloseHandle(hDiskFile);
	_ASSERT(dwFileSize == dwRead);

	if (Disk[unit] != NULL)
		HeapFree(hHeap, 0, Disk[unit]);
	Disk[unit] = pDisk;
	Status[unit] = 0x0001;			// disk present

	lstrcpy(Chipset.D70x_disk[unit], szFilename);

	Chipset.annun |= (1 << (3 + unit*2));
	DrawAnnunciator(4 + unit*2, TRUE);

	return TRUE;
}

BOOL hp9121_save(BYTE unit, LPCTSTR szFilename)
{
	HANDLE  hDiskFile = NULL;
	DWORD dwWritten;

	if (Disk[unit] == NULL)
		return FALSE;

	hDiskFile = CreateFile(szFilename,
						   GENERIC_WRITE,
						   FILE_SHARE_READ,
						   NULL,
						   CREATE_ALWAYS,
						   FILE_FLAG_SEQUENTIAL_SCAN,
						   NULL);
	if (hDiskFile == INVALID_HANDLE_VALUE)
	{
		hDiskFile = NULL;
		return FALSE;
	}

	// save file content
	WriteFile(hDiskFile, Disk[unit], DISK_BYTES, &dwWritten,NULL);
	CloseHandle(hDiskFile);
	lstrcpy(Chipset.D70x_disk[unit], szFilename);

	_ASSERT((DWORD)(DISK_BYTES) == dwWritten);

	return TRUE;
}

BOOL hp9121_eject(BYTE unit)
{
	if (Disk[unit] != NULL)
		HeapFree(hHeap, 0, Disk[unit]);
	Chipset.D70x_disk[unit][0] = 0x00;
	Disk[unit] = NULL;
	Status[unit] = 0x0000;			// no disk

	Chipset.annun &= ~(1 << (3 + unit*2));
	DrawAnnunciator(4 + unit*2, FALSE);
	
	return TRUE;
}

BOOL hp9121_widle(VOID)
{
	DWORD dwRefTime;

	SuspendDebugger();									// suspend debugger

	dwRefTime = timeGetTime();
	Chipset.idle = FALSE;
	Sleep(10);
	// wait for the Iddle state with 5 sec timeout
	while (((timeGetTime() - dwRefTime) < 5000L) && !(st9121 == 1))
		Sleep(10);

	if (st9121 == 1)									// not timeout, hp9121 is idle
		return FALSE;									// idle state
	else
		ResumeDebugger();								// timeout, resume to debugger

	return TRUE;										// hp9121 was busy
}

VOID hp9121_init(VOID)
{
	BYTE i;

	switch(Chipset.D70x_type)
	{
		case 0:
			hp9121_type[0] = 1; 
			hp9121_type[1] = 4;
			hp9121_config.heads = 2;
			hp9121_config.cylinders = 33;
			hp9121_config.sectors = 16;
			break;
		case 1:
			hp9121_type[0] = 0; 
			hp9121_type[1] = 129;
			hp9121_config.heads = 2;
			hp9121_config.cylinders = 75;
			hp9121_config.sectors = 30;
			break;
		case 2:
			hp9121_type[0] = 1; 
			hp9121_type[1] = 6;
			hp9121_config.heads = 4;
			hp9121_config.cylinders = 153;
			hp9121_config.sectors = 31;
			break;
		default:
			_ASSERT(0);
	}

	for (i = 0; i < 10; i++)
	{
		if ((Disk[i] == NULL) && (Chipset.D70x_disk[i][0] != 0x00))
		{
			hp9121_load(i, Chipset.D70x_disk[i]);
			// _ASSERT(Disk[i] != NULL);
		}
		if ((Disk[i] != NULL) && (Chipset.D70x_disk[i][0] != 0x00))
		{
			Chipset.annun |= (1 << (3 + i*2));
			DrawAnnunciator(4 + i*2, TRUE);
		}
		hp9121_head[i] = 0;
		hp9121_cylinder[i] = 0;
		hp9121_sector[i] = 0;
		hp9121_adr[i] = 0;
	}
	hp9121_unit = 0;
	st9121 = 0;								// state of hp9121 controller
	hp9121_untalk = FALSE;					// previous command was UNTALK ?
	hp9121_talk = FALSE;					// MTA received ?
	hp9121_listen = FALSE;					// MLA received ?
	hp9121_ppol_e = TRUE;					// parallel poll enabled
	hc_9121_hi = 0;							// hi mark
    hc_9121_lo = 0;							// lo mark
	hd_9121_hi = 0;							// hi mark
	hd_9121_lo = 0;							// lo mark
	hp9121_dsj = 0;
	hp9121_s1[0] = 0;
	hp9121_s1[1] = 0;
	hp9121_s2[0] = 0;
	hp9121_s2[1] = 0;
}

VOID hp9121_stop(VOID)
{
	int i;

	for (i = 0; i < 10; i++)
	{
		if (Disk[i] != NULL)
			HeapFree(hHeap, 0, Disk[i]);
	}
}
