/*
 *   Mops.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   Low level chipset, bus and CPU stuff based on docs
//
//   "Assembler ROM manuel" HP-87 Number 00087-90140
//   Assembler_ROM_Manual.pdf from Andr Koppel
//   "Assembler ROM manuel" HP-85 Number 00085-90944 Rev. B
//   HP85-Assembler.pdf from Vassili Prevelakis
//   "Original system rom listing from HP"
//   many pdf's from Andr Koppel
//
//   IO access are done in a normal way except for PTR access (Extended Memory Controller)
//   for an access at PTR (or PTR+, PTR-, PTR-+, with some minor exceptions):
//   an access in D mode is generally done to read 3 bytes (the value of the PTR)
//      so when when the READ demand 3 bytes at PTR, READIO send the value of PTR
//   an access in I mode is done in 2 memory access (see opcodes.c)
//   - the first to get the real address so the READ demand 2 bytes at PTR 
//   - the second use those 2 bytes as a new address for another READ
//     so when the READ demand 2 bytes at PTR, READIO send a new IO address (generally PTR + 0x10)
//     then READ demand s bytes at PTR + 0x10, READIO send the s bytes from the RAM
//
//   This way is consistent with I and D mode access at the normal RAM.
//
//   For WRITE it's more or less the same mecanism in reverse.
//
//   BUT it's entirely my own way to implement the EMC, perhaps it's completly
//   false. Actually I don't found any program which doesn't work with this.
//   Just an hint, it seem that the ASSEMBLER ROM use access in a way consistant 
//   with that (see around 0x61DE)
//
//   BUT p-System is quite strange and make some strange WRITE access at the EMC (see WriteIO)
//   perhaps that's why it doesn't work.
//   Ok there is a problem with that (0xB936: LDMD R22,=PTR1 to read the lower 16 bits of PTR1)
//
//   Another way could be to use byLaspOp to deal with that.
//   (all access in D mode have opcode 0xXY with Y < 8, Y > 7 for access in I mode)
//   Done and now p-SYSTEM works (03/08/04)
//

#include "StdAfx.h"
#include "HP8x.h"
#include "kml.h"
#include "io.h"
#include "mops.h"

//#define DEBUG_IO							// switch for I/O debug purpose
//#define DEBUG_CRT_IO						// switch for crt I/O debug purpose
//#define DEBUG_NO_CRT_IO					// switch for crt I/O debug purpose
//#define DEBUG_XMC_IO						// switch for emc I/O debug purpose
//#define DEBUG_NO_XMC_IO					// switch for emc I/O debug purpose
//#define DEBUG_KBD_IO						// switch for kbd I/O debug purpose
//#define DEBUG_CLK_IO						// switch for clk I/O debug purpose
//#define DEBUG_IOP_IO						// switch for IOP I/O debug purpose
//#define DEBUG_OTHER_IO					// switch for other I/O debug purpose

//#define DEBUG_TAPE_IO						// switch for TAPE I/O debug purpose
//#define DEBUG_PRT_IO

#if defined HP87
#define CRT_DISP 827
#endif
#if defined HP85
#define CRT_DISP 2268
#endif

#if defined DEBUG_IO

#if defined HP87

static WORD IO_ADR[] = {
	0xFF00,0xFF01,0xFF02,0xFF03,
	0xFF0A,0xFF0B,0xFF10,0xFF11,
	0xFF12,0xFF13,0xFF14,0xFF15,
	0xFF16,0xFF18,0xFF40,
//	0xFF4A,
//	0xFF4D,
	0xFF4A,0xFF4B,
	0xFF4C,0xFF4D,0xFF4E,0xFF4F,
	0xFF50,0xFF51,0xFF52,0xFF53,
	0xFF54,0xFF55,0xFF56,0xFF57,
	0xFF58,0xFF59,0xFF5A,0xFF5B,
	0xFF5C,0xFF5D,0xFF5E,0xFF5F,
	0xFF60,
	0xFFC0,0xFFC1,0xFFC2,0xFFC3,
	0xFFC4,0xFFC7,0xFFC8,0xFFC9,
	0xFFCA,0xFFCB,0xFFCC,0xFFCD,
	0xFFCE,0xFFCF
};

static TCHAR *IO_TXT[] = {
	_T("GINTEN"),_T("GINTDS"),_T("KEYSTS"),_T("KEYCOD"),
	_T("CLKSTS"),_T("CLKDAT"),_T("IOSTAT"),_T("IOINTC"),
	_T("IODATA"),_T("PPOLL"),_T("MLAD"),_T("SERPOL"),
	_T("EXSTAT"),_T("RSELEC"),_T("INTRSC"),
//	_T("ACTB-6"),
//	_T("ACTB-3"),
	_T("CCR/PSR00"),_T("OB/IB00"),
	_T("CCR/PSR01"),_T("OB/IB01"),_T("CCR/PSR02"),_T("OB/IB02"),
	_T("CCR/PSR03"),_T("OB/IB03"),_T("CCR/PSR04"),_T("OB/IB04"),
	_T("CCR/PSR05"),_T("OB/IB05"),_T("CCR/PSR06"),_T("OB/IB06"),
	_T("CCR/PSR07"),_T("OB/IB07"),_T("CCR/PSR08"),_T("OB/IB08"),
	_T("CCR/PSR09"),_T("OB/IB09"),_T("CCR/PSR10"),_T("OB/IB10"),
	_T("SC10+1"),
	_T("CRTSAD"),_T("CRTBAD"),_T("CRTSTS"),_T("CRTDAT"),
	_T("RULITE"),_T("SYSDIS"),_T("PTR1"),_T("PTR1-"),
	_T("PTR1+"),_T("PTR1-+"),_T("PTR2"),_T("PTR2-"),
	_T("PTR2+"),_T("PTR2-+")
};

#endif
#if defined HP85

static WORD IO_ADR[] = {
	0xFF00,0xFF01,0xFF02,0xFF03,
	0xFF04,0xFF05,0xFF06,0xFF07,
	0xFF08,0xFF09,
	0xFF0A,0xFF0B,
	0xFF0C,0xFF0D,0xFF0E,0xFF0F,
//	0xFF10,0xFF11,
//	0xFF12,0xFF13,0xFF14,0xFF15,
//	0xFF16,
	0xFF18,
	0xFF40,
//	0xFF4A,
//	0xFF4D,
	0xFF4A,0xFF4B,
	0xFF4C,0xFF4D,0xFF4E,0xFF4F,
	0xFF50,0xFF51,0xFF52,0xFF53,
	0xFF54,0xFF55,0xFF56,0xFF57,
	0xFF58,0xFF59,0xFF5A,0xFF5B,
	0xFF5C,0xFF5D,0xFF5E,0xFF5F,
	0xFF60,
//	0xFFC0,0xFFC1,0xFFC2,0xFFC3,
	0xFFC4,
//	0xFFC7,
	0xFFC8,0xFFC9,
	0xFFCA,0xFFCB,0xFFCC,0xFFCD,
	0xFFCE,0xFFCF
};

static TCHAR *IO_TXT[] = {
	_T("GINTEN"),_T("GINTDS"),_T("KEYSTS"),_T("KEYCOD"),
	_T("CRTSAD"),_T("CRTBAD"),_T("CRTSTS"),_T("CRTDAT"),
	_T("TAPESTS"),_T("TAPEDAT"),
	_T("CLKSTS"),_T("CLKDAT"),
	_T("PRT??"),_T("PRTDAT"),_T("PRTSTS"),_T("PRT!!"),
//	_T("IOSTAT"),_T("IOINTC"),
//	_T("IODATA"),_T("PPOLL"),_T("MLAD"),_T("SERPOL"),
//	_T("EXSTAT"),
	_T("RSELEC"),
	_T("INTRSC"),
//	_T("ACTB-6"),
//	_T("ACTB-3"),
	_T("CCR/PSR00"),_T("OB/IB00"),
	_T("CCR/PSR01"),_T("OB/IB01"),_T("CCR/PSR02"),_T("OB/IB02"),
	_T("CCR/PSR03"),_T("OB/IB03"),_T("CCR/PSR04"),_T("OB/IB04"),
	_T("CCR/PSR05"),_T("OB/IB05"),_T("CCR/PSR06"),_T("OB/IB06"),
	_T("CCR/PSR07"),_T("OB/IB07"),_T("CCR/PSR08"),_T("OB/IB08"),
	_T("CCR/PSR09"),_T("OB/IB09"),_T("CCR/PSR10"),_T("OB/IB10"),
	_T("SC10+1"),
//	_T("CRTSAD"),_T("CRTBAD"),_T("CRTSTS"),_T("CRTDAT"),
	_T("RULITE"),
//	_T("SYSDIS"),
	_T("PTR1"),_T("PTR1-"),
	_T("PTR1+"),_T("PTR1-+"),_T("PTR2"),_T("PTR2-"),
	_T("PTR2+"),_T("PTR2-+")
};

#endif
 
#define NLAB (sizeof(IO_ADR)/sizeof(IO_ADR[0]))

static TCHAR *IO_LAB(WORD d)
{
	WORD i;

	for(i = 0; i < NLAB; i++)
		if (IO_ADR[i] == d)
			return IO_TXT[i];
	return NULL;
}
#endif

//################
//#
//#    Public functions
//#
//################

VOID CpuReset(VOID)							// register setting after Cpu Reset
{
	int i;

	ZeroMemory(Chipset.R,sizeof(Chipset.R));
	Nread_nio(Chipset.R+4, 0x0000, 2);
	Chipset.ARP = 0;
	Chipset.DRP = 0;
	Chipset.E = 0;
	Chipset.DCM = FALSE;
	Chipset.CY = FALSE;
	Chipset.OVF = FALSE;
	Chipset.LSB = FALSE;
	Chipset.MSB = FALSE;
	Chipset.Z = FALSE;
	Chipset.LDZ = FALSE;
	Chipset.RDZ = FALSE;

	Chipset.cycles = 0x00000000;
	Chipset.cycles_reserved = 0x00000000;
	Chipset.dcycles = 0x0000;

	Chipset.Rom3 = 0;
	Chipset.RomCrc = 0x0000;
	ZeroMemory(Chipset.IORam,sizeof(Chipset.IORam));
	ZeroMemory(Chipset.Ram, Chipset.RamSize);
	
	Chipset.Ptr1 = 0x008000;
	Chipset.Ptr2 = 0x008000;

	ZeroMemory(Chipset.Video,Chipset.VideoSize);
	Chipset.crtsad = 0x0000;
	Chipset.crtoldsad = 0x0000;
	Chipset.crtbad = 0x0000;
	Chipset.crttimer = 0x00000000;
	Chipset.crtold = 0x00;
#if defined HP85
	Chipset.crtcount = 0x00;
#endif
	Chipset.keycode = 0;
	Chipset.keysts = 0x00;
	Chipset.keyenable = FALSE;
	Chipset.keyscanner = FALSE;
	Chipset.intkbd = FALSE;
	Chipset.KbdShift = FALSE;
	Chipset.KbdCtrl = FALSE;
	Chipset.KbdCaps = FALSE;
	Chipset.KbdDown = FALSE;

	ZeroMemory(Chipset.ptimer, 4 * sizeof(Chipset.ptimer[0]));
	ZeroMemory(Chipset.timer, 4 * sizeof(Chipset.timer[0]));
	ZeroMemory(Chipset.dtimer, 4 * sizeof(Chipset.dtimer[0]));
	Chipset.etimer[0] = Chipset.etimer[1] = FALSE;
	Chipset.etimer[2] = Chipset.etimer[3] = FALSE;
	Chipset.ctimer[0] = Chipset.ctimer[1] = FALSE;
	Chipset.ctimer[2] = Chipset.ctimer[3] = FALSE;
	Chipset.inttimer[0] = Chipset.inttimer[1] = FALSE;
	Chipset.inttimer[2] = Chipset.inttimer[3] = FALSE;
	Chipset.clksts = 0x00;
	Chipset.timern = 0x00;
	Chipset.clkaccess = FALSE;

	Chipset.intrsc = 0x00;
	for(i = 0; i < 11; i++)
	{
		int j = Chipset.iop[i].iop;

		ZeroMemory(&Chipset.iop[i], sizeof(Chipset.iop[0]));
		Chipset.iop[i].iop = j;
	}

	IopReload();				// initialize IO processors
	Chipset.inte = FALSE;		// no interrupt ??
	UpdateMainDisplay(TRUE);	// refresh screen
	UpdateAnnunciators();
}

////////////////////////////////////////////////////////////////////////////////
//
//   Bus Commands: for Capricorn, only strange mem access, it emit an address to select
//                 as chip then read/write as much byte as it wants .. 
//                 No info about chip-crossing access, perhaps some bugs due to that ?
//                 Actually chip crossing is permitted (use of do while structure)
//
//   Only IO here
//
////////////////////////////////////////////////////////////////////////////////
//
// read s bytes at d address in IO space
// IO access, gives 0xFF if no IO 
//
VOID ReadIO(BYTE *a, BYTE d, BYTE s)
{
	DWORD ptr;
	BYTE iop;

	switch (d)
	{
//	case 0x0000:	// GINTEN	0x0000	// GLOBAL INTERRUPT ENABLE
//		_ASSERT(0);
//		break;
//	case 0x0001:	// GINTDS	0x0001	// GLOBAL INTERRUPT DISABLE
//		_ASSERT(0);
//		break;
	case 0x0002:	// KEYSTS	0x0002	// KEYBOARD STATUS I/O ROUTINE
		_ASSERT(s == 1);
		a[0] = 0x00;
		if (Chipset.keyenable)					// keyboard enable
			a[0] |= 0x01;
		if (Chipset.KbdDown)					// key pressed
			a[0] |= 0x02;
		if (Chipset.KbdShift)					// shift pressed
			a[0] |= 0x08;
		if (Chipset.inte)						// global enable
			a[0] |= 0x80;
		break;
	case 0x0003:	// KEYCOD	0x0003	// KEYBOARD CODE AND EOJOB I/O ADDRESS
		_ASSERT(s == 1);
		a[0] = Chipset.keycode;
		break;

#if defined HP85

	case 0x0004:	// CRTSAD	0x0004	// CRT START ADDRESS I/O ADDRESS AS NIBBLES
		_ASSERT (s == 2);
		a[0] = (BYTE)(Chipset.crtsad & 0x00FF);
		a[1] = (BYTE)((Chipset.crtsad >> 8) & 0x3F);
		break;
	case 0x0005:	// CRTBAD	0x0005	// CRT NIBBLE ADDRESS
		_ASSERT (s == 2);
		a[0] = (BYTE)(Chipset.crtbad & 0x00FF);
		a[1] = (BYTE)((Chipset.crtbad >> 8) & 0x3F);
		break;
	case 0x0006:	// CRTSTS	0x0006	// CRT STATUS I/O ADDRESS
		_ASSERT (s == 1);
		if (Chipset.crttimer >= CRT_DISP )
			Chipset.crtrsts |= 0x02;	// display time
		else
			Chipset.crtrsts &= 0xFD;	// retrace time
		a[0] = Chipset.crtrsts | 0x60; 
		break;
	case 0x0007:	// CRTDAT	0x0007	// CRT DATA
		_ASSERT (s == 1);
		if ((Chipset.crtbad & 0x01) == 0x00)
		{
			a[0] = Chipset.Video[Chipset.crtbad >> 1];
			Chipset.crtbad++;
			Chipset.crtbad++;
		}
		else
		{
			a[0] = (Chipset.Video[Chipset.crtbad >> 1] & 0x0F) << 4;
			Chipset.crtbad++;
			if ((Chipset.crtwsts & 0x80) == 0x00)		// alpha mode	
			{
				if (Chipset.crtbad >= 0x1000)
					Chipset.crtbad -= 0x1000;
			}
			else
			{
				if (Chipset.crtbad >= 0x4000)
					Chipset.crtbad -= 0x4000;
			}

			a[0] |= (Chipset.Video[Chipset.crtbad >> 1] >> 4);
			Chipset.crtbad++;
		}
		if ((Chipset.crtwsts & 0x80) == 0x00)		// alpha mode	
		{
			if (Chipset.crtbad >= 0x1000)
				Chipset.crtbad -= 0x1000;
		}
		Chipset.crtbad &= 0x3FFF;
		Chipset.crtrsts &= 0x7E;						// crt not busy & no data
		break;
	case 0x0008:	//       	0x0008	// TAPESTS
		_ASSERT(s == 1);
		a[0] = 0xA0;
		break;
	case 0x0009:	//      	0x0009	// TAPEDAT
		_ASSERT(s == 1);
		a[0] = 0x00;
		break;

#endif
	
	case 0x000A:	// CLKSTS	0x000A	// CLOCK STATUS
		_ASSERT(s == 1);
		a[0] = ((!Chipset.clkaccess)<<7)|(Chipset.etimer[3]<<3)|(Chipset.etimer[2]<<2)|(Chipset.etimer[1]<<1)|(Chipset.etimer[0]);
		Chipset.clkaccess = FALSE;
		break;
	case 0x000B:	// CLKDAT	0x000B	// CLOCK DATA
		_ASSERT(s == 4);
		ptr = Chipset.timer[Chipset.timern];
		a[0] =  (BYTE)(ptr % 10);
		ptr /= 10;
		a[0] |= (BYTE)((ptr % 10) << 4);
		ptr /= 10;
		a[1] =  (BYTE)(ptr % 10);
		ptr /= 10;
		a[1] |= (BYTE)((ptr % 10) << 4);
		ptr /= 10;
		a[2] =  (BYTE)(ptr % 10);
		ptr /= 10;
		a[2] |= (BYTE)((ptr % 10) << 4);
		ptr /= 10;
		a[3] =  (BYTE)(ptr % 10);
		ptr /= 10;
		a[3] |= (BYTE)((ptr % 10) << 4);
		break;

#if defined HP85

	case 0x000C:	// PRT??
		_ASSERT(0);
		break;
	case 0x000D:	// PRTDAT ??
		_ASSERT(s == 1);
		a[0] = Chipset.prtdat;
		break;
	case 0x000E:	// PRTSTS ??
		_ASSERT(s == 1);
		a[0] = 0xc1;	// 0x80: paper in 0x40: printer ok ? 0x01: idle
		break;
	case 0x000F:	// PRT!!
		_ASSERT(0);
		break;

#endif

//	case 0x0010:	// IOSTAT	0x0010	// IO STATUS
//		_ASSERT(0);
//		break;
//	case 0x0011:	// IOINTC	0x0011	// IO CONTR-INTRUPT
//		_ASSERT(0);
//		break;
//	case 0x0012:	// IODATA	0x0012	// IO DATA
//		_ASSERT(0);
//		break;
//	case 0x0013:	// PPOLL	0x0013	// PARALLEL POLL REG
//		_ASSERT(0);
//		break;
//	case 0x0014:	// MLAD		0x0014	// SERIAL POLL REGISTER
//		_ASSERT(0);
//		break;
//	case 0x0015:	// SERPOL	0x0015	// MY LISTEN ADDRESS
//		_ASSERT(0);
//		break;
//	case 0x0016:	// EXSTAT	0x0016	// EXTENDED IO STATUS
//		_ASSERT(0);
//		break;
	case 0x0018:	// RSELEC	0x0018	// BANK SELECTABLE ROM SELECTION ADDRESS
		_ASSERT(s == 1);
		a[0] = Chipset.Rom3;
		break;
	case 0x0040:	// INTRSC	0x0040	// I/O CARDS SELECT CODE ADDRESS
		_ASSERT(s == 1);
		a[0] = Chipset.intrsc;
		break;
//	case 0x0041:	// lower IOP's not used normally
//	case 0x0042:
//	case 0x0043:
//	case 0x0044:
//	case 0x0045:
//	case 0x0046:
//	case 0x0047:
//	case 0x0048:
//	case 0x0049:
//		_ASSERT(0);
//		a[0] = 0xFF;
//		break;
	case 0x004A:	// ACTB-6	0x004A	// I/O MODULE ADDRESSES
	case 0x004C:
	case 0x004E:
	case 0x0050:	// ACTBAS	0x0050	// I/O MODULE ADDRESSES
	case 0x0052:
	case 0x0054:
	case 0x0056:
	case 0x0058:
	case 0x005A:
	case 0x005C:
	case 0x005E:
		_ASSERT(s == 1);
		iop = (d - 0x004A) >> 1;
		if (Chipset.iop[iop].iop)
		{
			a[0] = Chipset.iop[iop].psr;
		}
		else
			a[0]= 0xFF;
		break;
	case 0x004B:
	case 0x004D:	// ACTB-3	0x004D	// I/O MODULE ADDRESSES
	case 0x004F:
	case 0x0051:	// ACTBS+	0x0051	// I/O MODULE ADDRESSES
	case 0x0053:
	case 0x0055:
	case 0x0057:
	case 0x0059:
	case 0x005B:
	case 0x005D:
	case 0x005F:
		_ASSERT(s == 1);
		iop = (d - 0x004B) >> 1;
		if (Chipset.iop[iop].iop)
		{
			a[0] = Chipset.iop[iop].ib;
			Chipset.iop[iop].psr &= 0xFE;		// clear IBF bit in psr
		}
		else
			a[0]= 0xFF;
		break;
//	case 0x0060:	// SC10+1	0x0060	// I/O CARD STUFF
//		_ASSERT(0);
//		a[0]= 0xFF;
//		break;

#if defined HP87

	case 0x00C0:	// CRTSAD	0x00C0	// CRT START ADDRESS I/O ADDRESS
		_ASSERT(s == 2);
		a[0] = (BYTE)(Chipset.crtsad & 0x00FF);
#if defined VRAM64K
		a[1] = (BYTE)(Chipset.crtsad >> 8);
#else
		a[1] = (BYTE)((Chipset.crtsad >> 8) & 0x3F);
#endif
		break;
	case 0x00C1:	// CRTBAD	0x00C1	// CRT BYTE ADDRESS
		_ASSERT(s == 2);
		a[0] = (BYTE)(Chipset.crtbad & 0x00FF);
#if defined VRAM64K
		a[1] = (BYTE)(Chipset.crtbad >> 8);
#else
		a[1] = (BYTE)((Chipset.crtbad >> 8) & 0x3F);
#endif
		break;
	case 0x00C2:	// CRTSTS	0x00C2	// CRT STATUS I/O ADDRESS
		_ASSERT(s == 1);
		if (Chipset.crttimer > CRT_DISP )
			Chipset.crtsts |= 0x10;	// retrace time
		else
			Chipset.crtsts &= 0xEF;	// display time
		a[0] = Chipset.crtsts; 
		break;
	case 0x00C3:	// CRTDAT	0x00C3	// CRT DATA
		_ASSERT(s == 1);
		a[0] = Chipset.Video[Chipset.crtbad];
		Chipset.crtbad++;							// then wrap (really ?)
		if ((Chipset.crtsts & 0x80) == 0x00)		// alpha mode	
		{
			if ((Chipset.crtsts & 0x40) == 0x40)		// alpha all
			{
				if (Chipset.crtbad == 0x3FC0)
					Chipset.crtbad = 0x0000;
			}
			else
			{
				if (Chipset.crtbad == 0x10E0)
					Chipset.crtbad = 0x0000;
			}
		}
#if !defined VRAM64K
		Chipset.crtbad &= 0x3FFF;
#endif
		Chipset.crtsts &= 0xFE;						// crt not busy
		break;

#endif

//	case 0x00C4:	// RULITE	0x00C4	// RUN LIGHT I/O ADDRESS
//		_ASSERT(0);
//		break;
//	case 0x00C7:	// SYSDIS	0x00C7	// SOS CARD ROM DISABLE ADDRESS
//		_ASSERT(0);
//		break;
	case 0x00C8:	// PTR1		0x00C8	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)						// 2 bytes 
			{
				a[0] = 0xD8;					// 0xFFD8: IO addresse for indirect access to the ram
				a[1] = 0xFF;
			}
			else							// other ???
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr1 >> 16);
			}
		}
		break;
	case 0x00D8:		// IO addresse for Indirect access
		ptr = Chipset.Ptr1 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00C9:	// PTR1-	0x00C9	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xD9;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr1 >> 16);
			}
		}
		break;
	case 0x00D9:
		Chipset.Ptr1 -= s;
		ptr = Chipset.Ptr1 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00CA:	// PTR1+	0x00CA	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xDA;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr1 >> 16);
			}
		}
		break;
	case 0x00DA:
		ptr = Chipset.Ptr1 - 0x8000;
		Chipset.Ptr1 += s;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00CB:	// PTR1-+	0x00CB	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT ((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xDB;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr1 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr1 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr1 >> 16);
			}
		}
		break;
	case 0x00DB:
		ptr = Chipset.Ptr1 - s - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00CC:	// PTR2		0x00CC	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xDC;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr2 >> 16);
			}
		}
		break;
	case 0x00DC:
		ptr = Chipset.Ptr2 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00CD:	// PTR2-	0x00CD	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xDD;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr2 >> 16);
			}
		}
		break;
	case 0x00DD:
		Chipset.Ptr2 -= s;
		ptr = Chipset.Ptr2 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00CE:	// PTR2+	0x00CE	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xDE;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr2 >> 16);
			}
		}
		break;
	case 0x00DE:
		ptr = Chipset.Ptr2 - 0x8000;
		Chipset.Ptr2 += s;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	case 0x00CF:	// PTR2-+	0x00CF	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 2) || (s == 3));
		if ((byLastOp & 0x0F) > 7)		// I access
		{
			if (s == 2)
			{
				a[0] = 0xDF;
				a[1] = 0xFF;
			}
			else
			{
				_ASSERT(0);
			}
		}
		else							// D access
		{
			if (s == 2)
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
			}
			else
			{
				a[0] = (BYTE)(Chipset.Ptr2 & 0x0000ff);
				a[1] = (BYTE)((Chipset.Ptr2 >> 8 ) & 0x00ff);
				a[2] = (BYTE)(Chipset.Ptr2 >> 16);
			}
		}
		break;
	case 0x00DF:
		ptr = Chipset.Ptr2 - s - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				*a++ = Chipset.Ram[ptr];
			else
				*a++ = 0xFF;
			ptr++;
		} while (--s);
		break;
	default: 
		do			// for buggy programs
		{
			*a = 0xFF;
			a++;
			d++;
		} while (--s);
		break;
	}
	#if defined DEBUG_IO
	{
		int i;
		unsigned long l;
		DWORD j;
		TCHAR buffer[256];
		TCHAR *TXT = IO_LAB((WORD)(0xFF00|d));
		WORD pc = Chipset.R[4]|(Chipset.R[5]<<8);
		BOOL done = FALSE;

		#if defined DEBUG_XMC_IO
		if (d >= 0xC8 && d <= 0xDF && (pc < 0x7E1C || pc > 0x7E31))
		{
			done = TRUE;
			#if !defined (DEBUG_NO_XMC_IO)
			if ((d & 0x0F) < 0x0C)
			{
				if (TXT == NULL)
					i = wsprintf(buffer,_T("%04X: IO read :     %04X(%06X),%u = "), pc, 0xFF00|d, Chipset.Ptr1, s);
				else
					i = wsprintf(buffer,_T("%04X: IO read : %-8s(%06X),%u = "), pc, TXT, Chipset.Ptr1, s);
			}
			else
			{
				if (TXT == NULL)
					i = wsprintf(buffer,_T("%04X: IO read :     %04X(%06X),%u = "), pc, 0xFF00|d, Chipset.Ptr2, s);
				else
					i = wsprintf(buffer,_T("%04X: IO read : %-8s(%06X),%u = "), pc, TXT, Chipset.Ptr2, s);
			}
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
			#endif
		}
		#endif
		#if defined DEBUG_CRT_IO
#if defined HP87
		if (d >= 0xC0 && d <= 0xC3)
#endif
#if defined HP85
		if (d >= 0x04 && d <= 0x07)
#endif
		{
			done = TRUE;
			#if !defined (DEBUG_NO_CRT_IO)
			if (TRUE)
			{
				if (TXT == NULL)
					i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = \n"), pc, 0xFF00|d, s);
				else
					i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = \n"), pc, TXT, s);
				buffer[i++] = 0;
				OutputDebugString(buffer);
#if defined DEBUG_ONDISK
				WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
#endif
			}
			#endif
		}
		#endif
		#if defined DEBUG_KBD_IO
		if (d >= 0x02 && d <= 0x03)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = \n"), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = \n"), pc, TXT, s);
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_CLK_IO
		if (d >= 0x0A && d <= 0x0B)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_IOP_IO
		if (d >= 0x40 && d <= 0x80)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_OTHER_IO
		if (done == FALSE)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = \n"), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = %02X\n"), pc, TXT, s, a[0]);
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_TAPE_IO
		if (d >= 0x08 && d <= 0x09)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = %02X\n"), pc, 0xFF00|d, s, a[0]);
			else
				i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = %02X\n"), pc, TXT, s, a[0]);
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_PRT_IO
		if (d >= 0x0C && d <= 0x0F)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO read :     %04X,%u = %02X\n"), pc, 0xFF00|d, s, a[0]);
			else
				i = wsprintf(buffer,_T("%04X: IO read : %-8s,%u = %02X\n"), pc, TXT, s, a[0]);
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
	}
	#endif
}

//
// write s bytes at d address in IO space
// IO access
//
VOID WriteIO(BYTE *a, BYTE d, BYTE s)
{
	DWORD ptr;
	BYTE iop;

	#if defined DEBUG_IO
	{
		TCHAR buffer[256];
		DWORD j;
		int   i;
		unsigned long l;
		TCHAR *TXT = IO_LAB((WORD)(0xFF00|d));
		WORD pc = Chipset.R[4]|(Chipset.R[5]<<8);
		BOOL done = FALSE;

		#if defined DEBUG_XMC_IO
		if (d >= 0xC8 && d <= 0xDF && (pc < 0x7E1C || pc > 0x7E31)) 
		{
			done = TRUE;
			#if !defined (DEBUG_NO_XMC_IO)
			if ((d & 0x0F) < 0x0C)
			{
				if (TXT == NULL)
					i = wsprintf(buffer,_T("%04X: IO write:     %04X(%06X),%u = "), pc, 0xFF00|d, Chipset.Ptr1, s);
				else
					i = wsprintf(buffer,_T("%04X: IO write: %-8s(%06X),%u = "), pc, TXT, Chipset.Ptr1, s);
			}
			else
			{
				if (TXT == NULL)
					i = wsprintf(buffer,_T("%04X: IO write:     %04X(%06X),%u = "), pc, 0xFF00|d, Chipset.Ptr2, s);
				else
					i = wsprintf(buffer,_T("%04X: IO write: %-8s(%06X),%u = "), pc, TXT, Chipset.Ptr2, s);
			}
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
			#endif
		}
		#endif
		#if defined DEBUG_CRT_IO
#if defined HP87
		if (d >= 0xC0 && d <= 0xC3)
#endif
#if defined HP85
		if (d >= 0x04 && d <= 0x07)
#endif
		{
			done = TRUE;
			#if !defined (DEBUG_NO_CRT_IO)
			if (TRUE)
			{
				if (TXT == NULL)
					i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
				else
					i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
				for (j = 0; j < s; j++)
				{
					buffer[i] = a[j] >> 4;
					if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
					buffer[i] += _T('0');
					i++;
					buffer[i] = a[j] & 0xF;
					if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
					buffer[i] += _T('0');
					i++;
				}
				buffer[i++] = _T('\n');
				buffer[i++] = 0;
				OutputDebugString(buffer);
#if defined DEBUG_ONDISK
				WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
#endif
			}
			#endif
		}
		#endif
		#if defined DEBUG_KBD_IO
		if (d >= 0x02 && d <= 0x03)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_CLK_IO
		if (d >= 0x0A && d <= 0x0B)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_IOP_IO
		if (d >= 0x40 && d <= 0x80)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (i-1), &l, NULL); 
#endif
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_OTHER_IO
		if (done == FALSE)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
			OutputDebugString(buffer);
		}
		#endif
		#if defined DEBUG_TAPE_IO
		if (d >= 0x08 && d <= 0x09)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
		#if defined DEBUG_PRT_IO
		if (d >= 0x0C && d <= 0x0F)
		{
			if (TXT == NULL)
				i = wsprintf(buffer,_T("%04X: IO write:     %04X,%u = "), pc, 0xFF00|d, s);
			else
				i = wsprintf(buffer,_T("%04X: IO write: %-8s,%u = "), pc, TXT, s);
			for (j = 0; j < s; j++)
			{
				buffer[i] = a[j] >> 4;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
				buffer[i] = a[j] & 0xF;
				if (buffer[i] > 9) buffer[i] += _T('A') - _T('9') - 1;
				buffer[i] += _T('0');
				i++;
			}
			buffer[i++] = _T('\n');
			buffer[i++] = 0;
			OutputDebugString(buffer);
			done = TRUE;
		}
		#endif
	}
	#endif

	switch (d)
	{
	case 0x0000:	// GINTEN	0x0000	// GLOBAL INTERRUPT ENABLE
		_ASSERT(s == 1);
		Chipset.inte = TRUE;
		break;
	case 0x0001:	// GINTDS	0x0001	// GLOBAL INTERRUPT DISABLE
		_ASSERT(s == 1);
		Chipset.inte = FALSE;
		break;
	case 0x0002:	// KEYSTS	0x0002	// KEYBOARD STATUS I/O ROUTINE
		_ASSERT(s == 1);
		Chipset.keysts = a[0];
		if ((a[0] & 0x01) == 0x01)
		{
			Chipset.keyscanner = TRUE;
			Chipset.keyenable = TRUE;
		}
		if ((a[0] & 0x02) == 0x02)
			Chipset.keyenable = FALSE;
		if (bSound)
		{
			if ((a[0] & 0x20) == 0x00)
			{
				dwSoundTicks = Chipset.cycles;
				if (bInSoundxx && Chipset.R[4] == 0x1F && Chipset.R[5] == 0x11)
					HaltWaveInit();
			}
			else
			{
				dwSoundTicks = Chipset.cycles - dwSoundTicks + 108;
				PlayWavexx(613062L / dwSoundTicks);
			}
			if ((a[0] & 0x40) == 0x40)
				PlayWave12();
			else if (bInSound12)
				HaltWaveInit();
		}
		break;
	case 0x0003:	// KEYCOD	0x0003	// KEYBOARD CODE AND EOJOB I/O ADDRESS
		_ASSERT(s == 1);
		if (a[0] == 0x01)
			Chipset.keyscanner = TRUE;
		break;

#if defined HP85

	case 0x0004:	// CRTSAD	0x0004	// CRT START ADDRESS I/O ADDRESS
		_ASSERT (s == 2);
		if (s == 2)
			Chipset.crtoldsad = Chipset.crtsad;
			Chipset.crtsad = ((a[0] | (a[1]<<8)) & 0x3FFF);
			if (Chipset.crtoldsad != Chipset.crtsad)
			{
//				Chipset.crtrsts |= 0x80;			// crt busy
				UpdateMainDisplay(FALSE);
//				Chipset.crtrsts &= 0x7F;			// crt not busy
			}
		break;
	case 0x0005:	// CRTBAD	0x0005	// CRT NIBBLE ADDRESS
		_ASSERT (s == 2);
		if (s == 2)
			Chipset.crtbad = (a[0] | (a[1]<<8)) & 0x3FFF;
		break;
	case 0x0006:	// CRTSTS	0x0006	// CRT STATUS I/O ADDRESS
		_ASSERT(s == 1);
		if (s == 1)
		{
			if ((a[0] & 0x01) == 0x01)
				Chipset.crtrsts |= 0x01;
			Chipset.crtold = Chipset.crtwsts;
			Chipset.crtwsts = a[0] & 0xFE;
			if ((Chipset.crtwsts ^ (Chipset.crtold & 0xFE)) != 0x00)
			{
				if ((Chipset.crtwsts & 0x02) == 0x02)
				{
					Chipset.crtrsts |= 0x80;			// crt busy
					Chipset.crtcount = 0x02;
				}
				if ((Chipset.crtwsts & 0x80) == 0x80)
				{
					Chipset.crtrsts |= 0x80;			// graph mode
					if (Chipset.crtcount != 0x02)
						Chipset.crtcount += 0x01;
				}
				if ((Chipset.crtwsts & 0x82) == 0x82)
				{
					Chipset.crtrsts |= 0x80;			// graph mode
						Chipset.crtcount = 0x01;
				}
				UpdateMainDisplay(TRUE);
//				if ((Chipset.crtwsts & 0x02) == 0x00)
//					Chipset.crtrsts &= 0x7F;			// crt not busy
				Chipset.crtwsts &= 0xFE;
			}
		}
		break;
	case 0x0007:	// CRTDAT	0x0007	// CRT DATA
		_ASSERT(s == 1);
		if (s == 1)
		{
			if ((Chipset.crtrsts & 0x80) == 0x00)
			{
				Chipset.crtrsts |= 0x80;						// crt busy
				if ((Chipset.crtbad & 0x01) == 0x01)
				{
					Chipset.Video[Chipset.crtbad >> 1] &= 0xF0;
					Chipset.Video[Chipset.crtbad >> 1] |= (a[0] & 0xF0) >> 4;
					WriteToMainDisplay(Chipset.Video[Chipset.crtbad >> 1], Chipset.crtbad >> 1);
					Chipset.crtbad++;	
					if ((Chipset.crtwsts & 0x80) == 0x00)		// alpha mode	
					{
						if (Chipset.crtbad >= 0x1000)
							Chipset.crtbad -= 0x1000;
					} 
					else
					{
						if (Chipset.crtbad >= 0x4000)
							Chipset.crtbad -= 0x4000;
					}
					Chipset.Video[Chipset.crtbad >> 1] &= 0x0F;
					Chipset.Video[Chipset.crtbad >> 1] |= (a[0] & 0x0F) << 4;
					WriteToMainDisplay(Chipset.Video[Chipset.crtbad >> 1], Chipset.crtbad >> 1);
					Chipset.crtbad++;
				}
				else
				{
					Chipset.Video[Chipset.crtbad >> 1] = a[0];
					WriteToMainDisplay(Chipset.Video[Chipset.crtbad >> 1], Chipset.crtbad >> 1);
					Chipset.crtbad++;	
					Chipset.crtbad++;	
				}
				if ((Chipset.crtwsts & 0x80) == 0x00)		// alpha mode	
				{
					if (Chipset.crtbad >= 0x1000)
						Chipset.crtbad -= 0x1000;
				}
				Chipset.crtbad &= 0x3FFF;
				Chipset.crtrsts &= 0x7F;						// crt not busy
			}
		}
		break;
	case 0x0008:	// TAPESTS
		_ASSERT(s == 1);
		break;
	case 0x0009:	// TAPECRT
		_ASSERT(s == 1);
		break;

#endif

	case 0x000A:	// CLKSTS	0x000A	// CLOCK STATUS
		_ASSERT(s == 1);
		Chipset.timern = a[0] >> 6;
		if ((a[0] & 0x01) == 0x01)						// disable timer
			Chipset.etimer[Chipset.timern] = FALSE;
		if ((a[0] & 0x02) == 0x02)						// enable timer
			Chipset.etimer[Chipset.timern] = TRUE;
		if ((a[0] & 0x04) == 0x04)						// stop timer
		{
			Chipset.dtimer[Chipset.timern] = 0;
			Chipset.ctimer[Chipset.timern] = FALSE;
		}
		if ((a[0] & 0x08) == 0x08)						// start timer
		{
			Chipset.dtimer[Chipset.timern] = 0;
			Chipset.ctimer[Chipset.timern] = TRUE;
		}
		if ((a[0] & 0x10) == 0x10)						// clear timer
		{
			Chipset.dtimer[Chipset.timern] = 0;
			Chipset.timer[Chipset.timern] = 0;
		}
		break;
	case 0x000B:	// CLKDAT	0x000B	// CLOCK DATA
		_ASSERT(s == 4);
		Chipset.ptimer[Chipset.timern] = (a[0] & 0x0F) + 10*(a[0] >> 4) + 100*(a[1] & 0x0F) + 1000*(a[1] >> 4) +
						10000*(a[2] & 0x0F) + 100000*(a[2] >> 4) + 1000000*(a[3] & 0x0F) + 10000000*(a[3] >> 4);
		Chipset.clkaccess = TRUE;
		break;

#if defined HP85

	case 0x000C:	// PRT??
		_ASSERT(s == 1);
		Prt85_limit(a[0]);
		break;
	case 0x000D:	// PRTDAT ??
		_ASSERT(s == 1);
		Chipset.prtdat = a[0];
		Chipset.prtptr = a[0]*8;
		break;
	case 0x000E:	// PRTSTS ??
		_ASSERT(s == 1);
		if ((a[0] & 0x20) == 0x20)	// get bitmap of font
		{
			Chipset.prtdat = Prt85_get_byte((BYTE)(a[0] & 0x07));
		}
		if ((a[0] & 0xC0) == 0x40)	// text mode ?? 
		{
			Prt85_mode(0);
		}
		if ((a[0] & 0xC0) == 0xC0)	// graphic mode for the next 192 bytes ?
		{
			Prt85_mode(1);
		}
		break;
	case 0x000F:	// PRT!!
		// _ASSERT(s == 8); service rom !!
		iop = 0;
		while (iop < s)
			Prt85_print(a[iop++]);
		break;

#endif

//	case 0x0010:	// IOSTAT	0x0010	// IO STATUS
//		_ASSERT(0);
//		break;
//	case 0x0011:	// IOINTC	0x0011	// IO CONTR-INTRUPT
//		_ASSERT(0);
//		break;
//	case 0x0012:	// IODATA	0x0012	// IO DATA
//		_ASSERT(0);
//		break;
//	case 0x0013:	// PPOLL	0x0013	// PARALLEL POLL REG
//		_ASSERT(0);
//		break;
//	case 0x0014:	// MLAD		0x0014	// SERIAL POLL REGISTER
//		_ASSERT(0);
//		break;
//	case 0x0015:	// SERPOL	0x0015	// MY LISTEN ADDRESS
//		_ASSERT(0);
//		break;
//	case 0x0016:	// EXSTAT	0x0016	// EXTENDED IO STATUS
//		_ASSERT(0);
//		break;
	case 0x0018:	// RSELEC	0x0018	// BANK SELECTABLE ROM SELECTION ADDRESS
		_ASSERT(s == 1);
		Chipset.Rom3 = a[0];
		break;
	case 0x0040:	// INTRSC	0x0040	// I/O CARDS SELECT CODE ADDRESS
		_ASSERT(s == 1);
		break;			// do nothing actually
//	case 0x0041:	// lower IOP's not used normally
//	case 0x0042:
//	case 0x0043:
//	case 0x0044:
//	case 0x0045:
//	case 0x0046:
//	case 0x0047:
//	case 0x0048:
//	case 0x0049:
//		_ASSERT(0);
//		break;
	case 0x004A:	// ACTB-6	0x004A	// I/O MODULE ADDRESSES
	case 0x004C:
	case 0x004E:
	case 0x0050:	// ACTBAS	0x0050	// I/O MODULE ADDRESSES
	case 0x0052:
	case 0x0054:
	case 0x0056:
	case 0x0058:
	case 0x005A:
	case 0x005C:
	case 0x005E:
		_ASSERT(s == 1);
		iop = (d - 0x004A) >> 1;
		if (Chipset.iop[iop].iop)
		{
			Chipset.iop[iop].ccr = a[0];
			Chipset.iop[iop].ccrwr = TRUE;
		}
		break;
	case 0x004B:
	case 0x004D:	// ACTB-3	0x004D	// I/O MODULE ADDRESSES
	case 0x004F:
	case 0x0051:	// ACTBS+	0x0051	// I/O MODULE ADDRESSES
	case 0x0053:
	case 0x0055:
	case 0x0057:
	case 0x0059:
	case 0x005B:
	case 0x005D:
	case 0x005F:
		_ASSERT(s == 1);
		iop = (d - 0x004B) >> 1;
		if (Chipset.iop[iop].iop)
		{
			Chipset.iop[iop].ob = a[0];
			Chipset.iop[iop].psr |= 0x80;		// set OBF bit in psr
		}
		break;
//	case 0x0060:	// SC10+1	0x0060	// I/O CARD STUFF
//		_ASSERT(0);
//		break;

#if defined HP87

	case 0x00C0:	// CRTSAD	0x00C0	// CRT START ADDRESS I/O ADDRESS
		_ASSERT (s == 2);
		Chipset.crtoldsad = Chipset.crtsad;					// CRT start address
#if defined VRAM64K
		Chipset.crtsad = a[0] | (a[1]<<8);
#else
		Chipset.crtsad = (a[0] | (a[1]<<8)) & 0x3FFF;
#endif
		if (Chipset.crtoldsad != Chipset.crtsad)			// is the new one different of the previous
		{
			Chipset.crtsts |= 0x01;			// crt busy
			UpdateMainDisplay(FALSE);							// force full screen update
			Chipset.crtsts &= 0xFE;			// crt not busy
		}
		break;
	case 0x00C1:	// CRTBAD	0x00C1	// CRT BYTE ADDRESS
		_ASSERT (s == 2);
#if defined VRAM64K
		Chipset.crtbad = a[0] | (a[1]<<8);
#else
		Chipset.crtbad = (a[0] | (a[1]<<8)) & 0x3FFF;
#endif
		break;
	case 0x00C2:	// CRTSTS	0x00C2	// CRT STATUS I/O ADDRESS
		_ASSERT(s == 1);
		Chipset.crtold = Chipset.crtsts;
		Chipset.crtsts = a[0] & 0xEE;
		if ((Chipset.crtsts ^ (Chipset.crtold & 0xEE)) != 0x00)
		{
			Chipset.crtsts |= 0x01;			// crt busy
			UpdateMainDisplay(TRUE);
			Chipset.crtsts &= 0xFE;			// crt not busy
		}
		break;
	case 0x00C3:	// CRTDAT	0x00C3	// CRT DATA
		_ASSERT(s == 1);
		Chipset.crtsts |= 0x01;						// crt busy
#if defined VRAM64K
		Chipset.Video[Chipset.crtbad] = a[0];
#else
		Chipset.Video[Chipset.crtbad & 0x3FFF] = a[0];
#endif
		WriteToMainDisplay(a[0], (DWORD)( 0x0000FFFFL & (DWORD)(Chipset.crtbad)));
		Chipset.crtbad++;
		if ((Chipset.crtsts & 0x80) == 0x00)		// alpha mode	
		{
			if ((Chipset.crtsts & 0x40) == 0x40)		// alpha all
			{
				if (Chipset.crtbad == 0x3FC0)
					Chipset.crtbad = 0x0000;
			}
			else
			{
				if (Chipset.crtbad == 0x10E0)
					Chipset.crtbad = 0x0000;
			}
		}
#if !defined VRAM64K
		Chipset.crtbad &= 0x3FFF;
#endif
		Chipset.crtsts &= 0xFE;						// crt not busy
		break;

#endif

	case 0x00C4:	// RULITE	0x00C4	// RUN LIGHT I/O ADDRESS
		_ASSERT(s == 1);
		if (a[0] == 0)
		{
			Chipset.annun &= ~0x00000001;
			DrawAnnunciator(1, FALSE);
		}
		else
		{
			Chipset.annun |= 0x00000001;
			DrawAnnunciator(1, TRUE);
		}
		break;
//	case 0x00C7:	// SYSDIS	0x00C7	// SOS CARD ROM DISABLE ADDRESS
//		_ASSERT(0);
//		break;
	case 0x00C8:	// PTR1		0x00C8	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)													// for p-SYSTEM
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (Chipset.Ptr1 & 0xFF0000);
		}
		else 
		{
			_ASSERT(0);
		}
		break;
	case 0x00D8:
		ptr = Chipset.Ptr1 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);		
		break;
	case 0x00C9:	// PTR1-	0x00C9	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (Chipset.Ptr1 & 0xFF0000);
		}
		else 
		{
			_ASSERT(0);
		}
		break;
	case 0x00D9:
		Chipset.Ptr1 -= s;
		ptr = Chipset.Ptr1 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);		
		break;
	case 0x00CA:	// PTR1+	0x00CA	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (Chipset.Ptr1 & 0xFF0000);
		}
		else 
		{
			_ASSERT(0);
		}
		break;
	case 0x00DA:
		ptr = Chipset.Ptr1 - 0x8000;
		Chipset.Ptr1 += s;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);		
		break;
	case 0x00CB:	// PTR1-+	0x00CB	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr1 = a[0] | (a[1] << 8) | (Chipset.Ptr1 & 0xFF0000);
		}
		else
		{
			_ASSERT(0);
		}
		break;
	case 0x00DB:
		ptr = Chipset.Ptr1 - s - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);		
		break;
	case 0x00CC:	// PTR2		0x00CC	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (Chipset.Ptr2 & 0xFF0000);
		}
		else
		{
			_ASSERT(0);
		}
		break;
	case 0x00DC:
		ptr = Chipset.Ptr2 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);
		break;
	case 0x00CD:	// PTR2-	0x00CD	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (Chipset.Ptr2 & 0xFF0000);
		}
		else
		{
			_ASSERT(0);
		}
		break;
	case 0x00DD:
		Chipset.Ptr2 -= s;
		ptr = Chipset.Ptr2 - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);
		break;
	case 0x00CE:	// PTR2+	0x00CE	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (Chipset.Ptr2 & 0xFF0000);
		}
		else
		{
			_ASSERT(0);
		}
		break;
	case 0x00DE:
		ptr = Chipset.Ptr2 - 0x8000;
		Chipset.Ptr2 += s;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);
		break;
	case 0x00CF:	// PTR2-+	0x00CF	// I/O ADDRESSES FOR EMC POINTERS
		_ASSERT((s == 3) || (s == 2));
		if ((byLastOp & 0x0F) < 8)
		{
			if (s == 3)
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (a[2] << 16);
			if (s == 2)	
				Chipset.Ptr2 = a[0] | (a[1] << 8) | (Chipset.Ptr2 & 0xFF0000);
		}
		else
		{
			_ASSERT(0);
		}
		break;
	case 0x00DF:
		ptr = Chipset.Ptr2 - s - 0x8000;
		do
		{
			if (ptr < Chipset.RamSize)
				Chipset.Ram[ptr] = *a;
			ptr++;
			a++;
		} while (--s);
		break;
	default:
//		_ASSERT(0);
		break;
	}
}
