/*
 *   Mops.h
 *
 *   Copyright 2004 Olivier De Smet
 */

//#define DEBUG_MEMACC						// switch for MEMORY access debug purpose

//
//   Bus commands
//

//################
//#
//#    Low level subroutines
//#
//################

//
// better mem function as we move only from from 1 to 8 bytes at once without any alignment!
//
static __forceinline VOID memcopy(BYTE *d, BYTE *s, UINT count)
{
	while (count--)
		*d++ = *s++;
}

static __forceinline VOID memzero(BYTE *a, UINT count)
{
	while (count --)
		*a++ = 0x00;
}

static __forceinline VOID memfill(BYTE *a, UINT count)
{
	while (count --)
		*a++ = 0xFF;
}

////////////////////////////////////////////////////////////////////////////////
//
//   Bus Commands: for Capricorn, only strange mem access, it emit an address to select
//                 as chip then read/write as much byte as it wants .. 
//                 No info about chip-crossing access, perhaps some bugs due to that ?
//                 Actually chip crossing is permitted (use of do while structure)
//
//   Non IO here
//
////////////////////////////////////////////////////////////////////////////////

//
// read s bytes at d address with 24 bits address access and wrap (only for debugger)
// no IO access
//
VOID __inline Nread24_nio(BYTE *a, DWORD d, DWORD s)
{
	DWORD v;
	DWORD c;

	do
	{
		c = s;
		if (d >= 0x8000 && d < (0x8000 + Chipset.RamSize))	// Ram space (0x008000 - 0x008000 + RamSize)
		{
			v = d - 0x8000;
			if (c > Chipset.RamSize - v)
				c = Chipset.RamSize - v;
			memcopy(a, Chipset.Ram + v, c);
		}
		else if (d < 0x6000)								// Rom space (0x00000 - 0x005FFF)
		{
			if (c > 0x6000 - d)
				c = 0x6000 - d;
			memcopy(a, pbyRom + d, c);
		}
		else												// Paged Rom space (0x006000 - 0x007FFF)
		{
			v = d - 0x6000;
			if (c > 0x2000 - v)
				c = 0x2000 - v;
			if (Chipset.RomPage[Chipset.Rom3])					// xrom used ?
				memcopy(a, pbyRomPage[Chipset.Rom3] + v, c);		// read from paged rom
			else
				memfill(a, c);										// read 0xFF if not mapped
		}
		if (s -= c) {a += c; d += c;}
		if (d >= (0x8000 + Chipset.RamSize))				// wrap over memory (for the mem window of the debugger)
			d -= 0x8000 + Chipset.RamSize;
	} while (s);
	return;
}

//
// read 1 byte at d address with 16 bits address access
// no IO access
//
BYTE  __forceinline Nread_nio1(WORD d)
{
	if (d >= 0x8000)										// Ram space (0x8000 - 0xFFFF)	
		return Chipset.Ram[d - 0x8000];
	else if (d < 0x6000)									// Rom space (0x0000 - 0x5FFF)
		return pbyRom[d];
	else													// Paged rom space (0x6000 - 0x7FFF)
	{
		if (Chipset.RomPage[Chipset.Rom3])
			return pbyRomPage[Chipset.Rom3][d - 0x6000];
		else
			return 0xFF;												// read 0xFF if not mapped
	}
}

//
//   read 2 bytes at d address with 16 bits address access
//   no IO access and NO wrap (only for D access and =A effective address)
//   normal code don't cross boundaries (systemRom, sroms, ram, xram, ...)
//   due to the way the Capricorn bus works.
//   otherwise it's a bug in the program (not in the emulator)
//
WORD  __forceinline Nread_nio2(WORD d)
{
		if (d >= 0x8000)									// Ram space (0x8000 - 0xFFFF)
			return (*((WORD *)(Chipset.Ram + d - 0x8000)));
		else if (d < 0x6000)								// Rom space (0x000 - 0x5FFF)
			return (*((WORD *)(pbyRom + d)));
		else												// Paged Rom space (0x6000 - 0x7FFF)
			if (Chipset.RomPage[Chipset.Rom3])
				return (*((WORD *)(pbyRomPage[Chipset.Rom3] + d - 0x6000)));
			else
				return 0xFFFF;
}

//
// read s bytes at d address with 16 bits address access
// no IO access
//
VOID __forceinline Nread_nio(BYTE *a, WORD d, WORD s)
{
	WORD v;
	WORD c;

	do
	{
		c = s;
		if (d >= 0x8000)									// Ram space (0x8000 - 0xFFFF)
		{
			v = d - 0x8000;
			if (c > 0x8000 - v)
				c = 0x8000 - v;
			memcopy(a, Chipset.Ram + v, c);
		}
		else if (d < 0x6000)								// Rom space (0x000 - 0x5FFF)
		{
			if (c > 0x6000 - d)
				c = 0x6000 - d;
			memcopy(a, pbyRom + d, c);
		}
		else												// Paged Rom space (0x6000 - 0x7FFF)
		{
			v = d - 0x6000;
			if (c > 0x2000 - v)
				c = 0x2000 - v;
			if (Chipset.RomPage[Chipset.Rom3])
				memcopy(a, pbyRomPage[Chipset.Rom3] + v, c);
			else
				memfill(a, c);									// read 0xFF if not mapped
		}
		if (s -= c) {a += c; d += c;}
	} while (s);
	return;
}

//
// read 1 byte at d address with 16 bits address access
// IO access
//
BYTE __forceinline Nread1(WORD d)
{
	#if defined DEBUG_MEMACC
	{
		TCHAR buffer[256];
		if (d >= (0x895B-8) && (d+s) <= (0x895C+8))
		{
			wsprintf(buffer,_T("%04X: Mem read : %04X,%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), d, 1);
			OutputDebugString(buffer);
		}
	}
	#endif

	if (d >= 0xFF00)								// I/O space (0xFF00 - 0xFFFF)
	{
		BYTE a;
		ReadIO(&a, (BYTE)(d), (BYTE)(1));
		return a;
	}
	else if (d >= 0x8000)							// Ram space (0x8000 - 0xFF00)
		return Chipset.Ram[d - 0x8000];
	else if (d < 0x6000)							// Rom space (0x0000 - 0x5FFF)
		return pbyRom[d];
	else											// Paged Rom space (0x6000 - 0x7FFF)
	{
		if (Chipset.RomPage[Chipset.Rom3])
			return pbyRomPage[Chipset.Rom3][d - 0x6000];
		else
			return 0xFF;									// read 0xFF is not mapped
	}
}

//
// read s bytes at d address with 16 bits address access
// IO access
//
VOID __forceinline Nread(BYTE *a, WORD d, WORD s)
{
	WORD v;
	WORD c;

	#if defined DEBUG_MEMACC
	{
		TCHAR buffer[256];
		if (d >= (0x895B-8) && (d+s) <= (0x895C+8))
		{
			wsprintf(buffer,_T("%04X: Mem read : %04X,%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), d, s);
			OutputDebugString(buffer);
		}
	}
	#endif

	do
	{
		c = s;
		if (d >= 0xFF00)							// I/O space (0xFF00 - 0xFFFF)
		{
			v = (BYTE)(d);
			if (c > 0x0100 - v)
				c = 0x0100 - v;
			ReadIO(a, (BYTE)(v), (BYTE)(c));
		}
		else if (d >= 0x8000)						// Ram space (0x8000 - 0xFEFF)
		{
			v = d - 0x8000;
			if (c > 0x7F00 - v)
				c = 0x7F00 - v;
			memcopy(a, Chipset.Ram + v, c);
		}
		else if (d < 0x6000)						// Rom space (0x0000 - 0x5FFF)
		{
			if (c > 0x6000 - d)
				c = 0x6000 - d;
			memcopy(a, pbyRom + d, c);
		}
		else										// Paged Rom space (0x6000 - 0x7FFF)
		{
			v = d - 0x6000;
			if (c > 0x2000 - v)
				c = 0x2000 - v;
			if (Chipset.RomPage[Chipset.Rom3])
				memcopy(a, pbyRomPage[Chipset.Rom3] + v, c);
			else
				memfill(a, c);							// read 0xFF if not mapped
		}
		if (s -= c) {a += c; d += c;}
	} while (s);
}

//
// write 1 byte at d address with 16 bits address access
// IO access
//
VOID __forceinline Nwrite1(BYTE a, WORD d)
{
	#if defined DEBUG_MEMACC
	{
		TCHAR buffer[256];
		if (d >= (0x895B-8) && (d+s) <= (0x895C+8))
		{
			wsprintf(buffer,_T("%04X: Mem write: %04X,%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), d, s);
			OutputDebugString(buffer);
		}
	}
	#endif

	if (d >= 0xFF00)									// I/O space (0xFF00 - 0xFFFF)
		WriteIO(&a, (BYTE)(d), (BYTE)(1));
	else if (d >= 0x8000)								// Ram space (0x8000 - 0xFEFF)
		Chipset.Ram[d - 0x8000] = a;
}

//
// write s bytes at d address with 16 bits address access
// IO access
//
VOID __forceinline Nwrite(BYTE *a, WORD d, WORD s)
{
	WORD v;
	WORD c;

	#if defined DEBUG_MEMACC
	{
		TCHAR buffer[256];
		if (d >= (0x895B-8) && (d+s) <= (0x895C+8))
		{
			wsprintf(buffer,_T("%04X: Mem write: %04X,%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), d, s);
			OutputDebugString(buffer);
		}
	}
	#endif

	do
	{
		c = s;
		if (d >= 0xFF00)								// I/O space (0xFF00 - 0xFFFF)
		{
			v = (BYTE)(d);
			if (c > 0x0100 - v)
				c = 0x0100 - v;
			WriteIO(a, (BYTE)(v), (BYTE)(c));
		}
		else
		{
			if (c > 0xFF00 - d)
				c = 0xFF00 - d;
			if (d >= 0x8000)							// Ram space (0x8000 - 0xFEFF)
				memcopy(Chipset.Ram + d - 0x8000, a, c);
		}
		if (s -= c) {a += c; d += c;}
	} while (s);
}

